#include <cstdio>
#include "coeff.h"

using namespace std;

const char *ReadableTypeName<Fpoint>::name() {
	static char str[50];
	sprintf(str,"Mp_float%ld",Fpoint::precision());
	return str;
}

const char *ReadableTypeName<Zp>::name() {
	static char str[50];
	sprintf(str,"Z_%d",Zp::mod());
	return str;
}

Coeff &Coeff::operator=(const Coeff &c) {
	if (m_data!=c.m_data) {
  	if (m_data) {
  		if (c.m_data) {
  			if (m_data->type()==c.m_data->type())
  				m_data->copy(c.m_data);
  			else {
  				m_data->destroy();
  				m_data = c.m_data->clone();
  			}
  		}
  		else {
  			m_data->destroy();
  			m_data=0;	
  		}
  	}
  	else
  	if (c.m_data) m_data=c.m_data->clone();
	}
	return *this;
}

bool Coeff::isLess(const Coeff &c) const {
  if (m_data==c.m_data) return false;
	if (!m_data || !c.m_data) return false;
	return m_data->isLess(c.m_data);	
}

bool Coeff::isEqual(const Coeff &c) const {
	if (m_data==c.m_data) return true;
	if (!m_data || !c.m_data) return false;
	return m_data->isEqual(c.m_data);
}

void Coeff::print(ostream &os) const {
	if (m_data) m_data->print(os);
	else os<<"<void>";
}

void Coeff::scan(istream &is) {
	data()->scan(is);
}

const type_info &Coeff::type() const {
	return m_data?m_data->type():SafeType<void>::id();
};

const char *Coeff::typeName() const {
	if (m_data) return m_data->typeName();
	return "void";
}

template class Tdomain<float>;
template class Tdomain<double>;
template class Tdomain<Rational>;
template class Tdomain<Fpoint>;
template class Tdomain<Zp>;

bool AlmostEqual::isEqual(const Coeff &c1, const Coeff &c2) const {
	static void *dispatch[]={&&l_exact, &&l_abs, &&l_rel};
	goto *dispatch[scheme()];
l_exact:
	return c1==c2;
l_abs:
	return (c2-c1).abs()<epsilon();
l_rel:
	return (c2-c1).abs()<max(c1.abs(),c2.abs())*epsilon();
}

bool AlmostEqual::isZero(const Coeff &c) const {
	if (scheme()==Exact) return c.isZero();
	return c.abs()<epsilon();
}

const char *AlmostEqual::schemeName(Scheme scm) {
	switch (scm) {
		case Exact:return "exact";
		case Abs:  return "abs";
		case Rel:  return "rel";
		default :  return "UNDEFINED";
	}
}

void AlmostEqual::print(ostream &os) const {
	const char *name=schemeName(scheme());
	switch (scheme()) {
		case Abs  :
		case Rel  :os<<name<<'<'<<epsilon();break;
		case Exact:
		default   :os<<name;
	}
}
