#if !defined(__EMATRIX_H__)
#define __EMATRIX_H__

#include <functional>
#include <algorithm>
#include <set>
#include "basis.h"
#include "report.h"

class Elimination: public Report, public Stat {

	typedef BasisBase::const_iterator basis_iterator;

  struct IterAddrLess: public std::binary_function<basis_iterator,basis_iterator,bool> {
		bool operator()(basis_iterator i1, basis_iterator i2) const {
			return &*i1<&*i2;
		}
	};

  struct BetterPivot {
  	bool operator()(const Polin &p1, const Polin &p2, int pind1, int pind2) const {
  		return isBetter(p1,p2,pind1,pind2);
  	}
  private:
    bool isBetter(const Polin &p1, const Polin &p2, int pind1, int pind2) const {
      if (p1.size()==1 && p2.size()> 1) return true;
      if (p1.size()> 1 && p2.size()==1) return false;
      const Monom &m1=p1[pind1];
      const Monom &m2=p2[pind2];
      double ac1=std::fabs(m1.coeff().toDouble());
      double ac2=std::fabs(m2.coeff().toDouble());
      if (ac2<ac1) return true;
      if (ac1<ac2) return false;
/*      if (p1.context().eto()==EtermOrder::te()) {
        int led1=p1.lediff();
        int led2=p2.lediff();
        if (led1<led2) return true;
        if (led2<led1) return false;
      }*/
      if (m1.eterm()<m2.eterm()) return true;
      if (m2.eterm()<m1.eterm()) return false;
      if (p1.size()<p2.size()) return true;
      if (p2.size()<p1.size()) return false;
      return p1.sdegree()<p2.sdegree();
#if 0       
      if (p1.sdegree()<p2.sdegree()) return true;
      if (p2.sdegree()<p1.sdegree()) return false;
      return (p1.size()<p2.size());
#endif       
    }
  };

/*	struct IterBetterPivot: public binary_function<basis_iterator,basis_iterator,bool> {
		bool operator()(basis_iterator i1, basis_iterator i2) {
			if (i1==i2) return false;
			if (m_better(*i1,*i2,m_et)) return true;
			if (m_better(*i2,*i1,m_et)) return false;
			return &*i1<&*i2;
		}
		IterBetterPivot(const Eterm &et):m_et(et) {}
	private:
		const Eterm m_et;
		BetterPivot m_better;
	};*/

	struct SetOfRows: public std::set<basis_iterator,IterAddrLess> {
		bool contains(basis_iterator bi) const {
			return find(bi)!=end();
		}
		void print(std::ostream &os) const {
			if (empty()) os<<"<empty>";
			else {
				for(const_iterator i=begin();i!=end();++i) os<<' '<<(**i).tag();
			}
		}
		friend inline std::ostream &operator<<(std::ostream &os, const SetOfRows &sor) {
			sor.print(os);
			return os;
		}
	};

  class pivot_iterator: public basis_iterator  {
    int m_pindex;
  public:
    pivot_iterator(basis_iterator bi, int pi): basis_iterator(bi),m_pindex(pi) {}
    const Monom &pm() const {return (*this)->monom(m_pindex);}
    int pindex() const {return m_pindex;}
  };

  struct SetOfPivots: public std::set<pivot_iterator,IterAddrLess> {
    bool contains(pivot_iterator bi) const {
			return find(bi)!=end();
		}
		void print(std::ostream &os) const {
			if (empty()) os<<"<empty>";
			else {
				for(const_iterator i=begin();i!=end();++i) os<<' '<<(*i).pm()<<':'<<(**i).tag();
			}
		}
		friend inline std::ostream &operator<<(std::ostream &os, const SetOfPivots &sop) {
			sop.print(os);
			return os;
		}
  };
    
	struct Cols: public std::map<Eterm,SetOfPivots> {
		typedef std::map<Eterm, SetOfPivots> Inherited;
		typedef Inherited::key_type key_type;
		typedef Inherited::mapped_type mapped_type;
		typedef mapped_type data_type;
		typedef Inherited::iterator iterator;
		typedef BasisBase::const_iterator basis_iterator;
		BasisBase &m_basis;

		typedef basis_iterator::value_type::const_iterator pol_iterator;

		void insert_cols(basis_iterator bi);
		void erase_cols(basis_iterator bi);
	public:

		typedef SetOfPivots::const_iterator col_iterator;

		Cols(BasisBase &bb):m_basis(bb) {}
		data_type &operator[](const key_type &key) {
			iterator pos=find(key);
			if (pos==end()) {
				std::pair<iterator,bool> result=Inherited::insert(value_type(key,SetOfPivots()));
				PRECONDITION(result.second);
				pos=result.first;
			}
			return pos->second;
		}
		const data_type &operator[](const key_type &key) const {
			const_iterator pos=find(key);
			PRECONDITION(pos!=end());
			return pos->second;
		}
		basis_iterator insert(BasisBase::const_reference bp) {
			basis_iterator bi=m_basis.insert(m_basis.end(),bp);
			insert_cols(bi);
			return bi;
		}
		void erase(basis_iterator bi) {
			erase_cols(bi);
			m_basis.erase(bi);
		}
		void add(const SetOfEterms &st);
		void print(std::ostream &os) const {
			if (empty()) os<<"<empty>\n";
			else {
				for(const_iterator i=begin();i!=end();++i) {
					os<<i->first;
					i->second.print(os);
					os<<'\n';
				}
			}
		}
		col_iterator col_begin(const Eterm &et) const {return (*this)[et].begin();}
		col_iterator col_end(const Eterm &et) const {return (*this)[et].end();}


		bool pivotR(col_iterator &result, const Eterm &et, const SetOfRows &forbidden) const {
			col_iterator first=col_begin(et);
			col_iterator  last=col_end(et);
      col_iterator pivot=last;
			for (;first!=last;++first) {
				if ( forbidden.contains(*first) ) continue;
				if ( pivot==last || m_betterPivot(**first,**pivot,first->pindex(),pivot->pindex()) ) pivot=first;
			}
      result=pivot;
      return pivot!=last;
		}

		bool pivotK(col_iterator &result, const Eterm &et) const {
			col_iterator first=col_begin(et);
			col_iterator  last=col_end(et);
      col_iterator pivot=last;
			for (;first!=last;++first) {
        if (first->pindex()!=0) continue;// if ((**first).lt() != et) continue;
				if ( pivot==last || m_betterPivot(**first,**pivot,first->pindex(),pivot->pindex()) ) pivot=first;
			}
      result=pivot;
      return pivot!=last;
		}

		bool pivotR(col_iterator &result, const SetOfEterms &rterms, const SetOfRows &forbidden) const {
			SetOfEterms::const_iterator first=rterms.begin();
			SetOfEterms::const_iterator  last=rterms.end();
			col_iterator pivot;
      bool havePivot=false;
			for (;first!=last;++first) {
        col_iterator pprobe;
        if ( !pivotR(pprobe,*first,forbidden) ) continue;
        PRECONDITION( pprobe->pm().eterm()==*first);
				if ( !havePivot || m_betterPivot(**pprobe,**pivot,pprobe->pindex(),pivot->pindex()) ) {
          pivot=pprobe;
          havePivot=true;
				}
			}
      result=pivot;
			return havePivot;
		}

		bool pivotK(col_iterator &result, const SetOfEterms &kterms) const {
			SetOfEterms::const_iterator first=kterms.begin();
			SetOfEterms::const_iterator  last=kterms.end();
			col_iterator pivot;
      bool havePivot=false;
			for (;first!=last;++first) {
        col_iterator pprobe;
        if ( !pivotK(pprobe,*first) ) continue;
        PRECONDITION( pprobe->pm().eterm()==*first);
				if ( !havePivot || m_betterPivot(**pprobe,**pivot,pprobe->pindex(),pivot->pindex()) ) {
					pivot=pprobe;
          havePivot=true;
				}
			}
      result=pivot;
			return havePivot;
		}
	private:
		BetterPivot m_betterPivot;
	};

public:

	typedef Cols::col_iterator col_iterator;
	typedef BasisBase Ematrix;
	typedef BasisBase::const_iterator ematrix_iterator;

private:

	Basis m_simp;

	void eliminate(Ematrix &em, const SetOfEterms &ired, const SetOfEterms &ikeep,  bool keePivots);
	
public:

	Elimination(const Report &r, const Stat &s):Report(r),Stat(s) {
		stat().add("elim","Number of row eliminations");
	}
	void operator()(Basis &b, const SetOfEterms &ired, const SetOfEterms &ikeep,  bool keePivots=true);
	
	void clear() {m_simp.clear();}
	const Basis &simp() const {return m_simp;}
};

#endif // __EMATRIX_H__
