#include <iostream>
#include "etermorder.h"
#include "eterm.h"

bool ExtOrder::equal(const Eterm &et1, const Eterm &et2) const {
	return (&et1==&et2) || (et1.edegree()==et2.edegree() && et1.term()==et2.term());
}

struct To: public ExtOrder {
	virtual bool less(const Eterm &et1, const Eterm &et2) const;
	virtual void print(std::ostream &os) const;
};

bool To::less(const Eterm &et1, const Eterm &et2) const {
	return &et1!=&et2 && et1.term()<et2.term();
}

void To::print(std::ostream &os) const {
	os<<"To";
}

struct Te: public ExtOrder {
	virtual bool less(const Eterm &et1, const Eterm &et2) const;
	virtual void print(std::ostream &os) const;
};

bool Te::less(const Eterm &et1, const Eterm &et2) const {
  if (&et1==&et2) return false;
	return et1.term()<et2.term() || (et1.term()==et2.term() && et2.edegree()<et1.edegree());
}

void Te::print(std::ostream &os) const {
	os<<"Te";
}

struct eT: public ExtOrder {
	virtual bool less(const Eterm &et1, const Eterm &et2) const;
	virtual void print(std::ostream &os) const;
};

bool eT::less(const Eterm &et1, const Eterm &et2) const {
  if (&et1==&et2) return false;
	return et2.edegree()<et1.edegree() || (et1.edegree()==et2.edegree() && et1.term()<et2.term());
}

void eT::print(std::ostream &os) const {
	os<<"eT";
}

EtermOrder EtermOrder::to() {
	static To order;
	return EtermOrder(&order);
}

EtermOrder EtermOrder::te() {
	static Te order;
	return EtermOrder(&order);
}

EtermOrder EtermOrder::et() {
	static eT order;
	return EtermOrder(&order);
}
