#include <cctype>
#include <iomanip>
#include <functional>
#include <algorithm>
#include <list>
#include "inputparser.h"
#include "report.h"

using namespace std;

void InputParser::Polins::print(ostream &os) const {
	copy(begin(),end(),ostream_iterator<Polin>(os,"\n"));
}

struct func_toupper: public unary_function<char, char> {
	char operator()(char ch) const {
		return std::toupper( ch );
  }
};

struct func_tolower: public unary_function<char, char> {
	char operator()(char ch) const {
		return std::tolower( ch );
  }
};


inline std::string uppercase( const std::string& str ){
	std::string result( str );
  std::transform( result.begin(), result.end(), result.begin(),func_toupper() );
	return result;
}

inline std::string lowercase( const std::string& str ){
	std::string result( str );
  std::transform( result.begin(), result.end(), result.begin(),func_tolower() );
	return result;
}

bool InputParser::operator()() {
	m_results.clear();
	for(;;) {
  	SlocString command=name();
  	if (command) {
			string commData=lowercase(command.data());
			if ( commData=="alphabet" ) {
				symb("=");
				list<string> vars;
				for (;;) {
  				SlocString var=name();
  				if (!var) {
  					error().set("variable name expected",var);
  					return false;
  				}
  				if ( find(vars.begin(),vars.end(),var.data())==vars.end() ) {
  					vars.push_back(var.data());
  				}
  				else {
  					error().set( string("variable name '")+var.data()+string("' is already defined"),var);
  					return false;
  				}
					if ( !symb(",") ) break;
				}
				m_results["alphabet"]=Alphabet(vars.begin(),vars.end());
			}
			else
			if ( commData=="termorder" ) {
				symb("=");
				SlocString order=name();
				if (order) {
					if (order.data()=="lex") m_results["termorder"]=TermOrder::lex();
					else
					if (order.data()=="revlex") m_results["termorder"]=TermOrder::revlex();
					else
					if (order.data()=="tdeg") m_results["termorder"]=TermOrder::tdeg();
					else
					if (order.data()=="revtdeg") m_results["termorder"]=TermOrder::revtdeg();
					else {
						error().set(order.data()+" is not a valid term order",order);
						return false;
					}
				}
				else {
					error().set(string("Term order expected but got '")+order.data()+"'",order);
					return false;
				}
      }
			else
			if ( commData=="extorder" ) {
				symb("=");
				SlocString extorder=name();
				if (extorder) {
					if (extorder.data()=="To") m_results["etermorder"]=EtermOrder::to();
					else
					if (extorder.data()=="Te") m_results["etermorder"]=EtermOrder::te();
					else
					if (extorder.data()=="eT") m_results["etermorder"]=EtermOrder::et();
					else {
						error().set(extorder.data()+" is not a valid term order extension",extorder);
						return false;
					}
				}
				else {
					error().set(string("Term order extension expected but got '")+extorder.data()+"'",extorder);
					return false;
				}
      }
			else
			if ( commData=="domain" ) {
				symb("=");
				SlocString domain=name();
				if (domain) {
					string domData=lowercase(domain.data());
					if (domData=="float") {
						m_results["domain"]=Coeff(float(1));
						int dig=7;
						SlocString extra=symb(",");
						if ( extra ) {
			 				IntExprParser<int> digParser( scanner(),error(),true, 0,1);
 							if (!digParser.expr(dig)) return false;
						}
						else {	
							cerr<<"The number of printed digits for float is not specified, using "<<dig<<" by default\n";
						}
						if (dig<2) {
							error().set("The number of printed digits for float must be >=2",extra);
							return false;
						}
						m_results["digits"]=dig;
					}
					else
					if (domData=="double") {
						m_results["domain"]=Coeff(double(1));
						int dig=15;
						SlocString extra=symb(",");
						if ( extra ) {
			 				IntExprParser<int> digParser( scanner(),error(),true, 0,1);
 							if (!digParser.expr(dig)) return false;
						}
						else {	
							cerr<<"The number of printed digits for double is not specified, using "<<dig<<" by default\n";
						}
						if (dig<2) {
							error().set("The number of printed digits for double must be >=2",extra);
							return false;
						}
						m_results["digits"]=dig;
					}
					else
					if (domData=="fpoint") {
						SlocString extra=symb(",");
						int prec=256;
						int dig=80;						
						if ( extra ) {
			 				IntExprParser<int> precParser( scanner(),error(),true, 0,1);							
 							if (!precParser.expr(prec)) return false;
						}
						else {	
							cerr<<"the length of mantissa is not specified, using "<<prec<<" by default\n";
						}
						if (prec<16) {
							error().set("the length of mantisa must be >=16",extra);
							return false;
						}
						extra=symb(",");
						if ( extra ) {
			 				IntExprParser<int> digParser( scanner(),error(),true, 0,1);
 							if (!digParser.expr(dig)) return false;
						}
						else {	
							cerr<<"The number of printed digits is not specified, using "<<dig<<" by default\n";
						}
						if (dig<2) {

							error().set("The number of printed digits must be >=2",extra);
							return false;
						}
            Fpoint::precision(prec);
						Fpoint::digits(dig);
						m_results["digits"]=dig;
						m_results["domain"]=Coeff(Fpoint(1));
					}
					else
					if (domData=="rational") m_results["domain"]=Coeff(Rational(1));
					else
					if (domData=="zp") {
						SlocString extra=symb(",");
						int mod=65521;
						if ( extra ) {
			 				IntExprParser<int> modParser( scanner(),error(),true, 0,1);							
 							if (!modParser.expr(mod)) return false;
						}
						else {	
							cerr<<"the characteristic not specified, using "<<mod<<" by default\n";
						}
 						if (mod<2) {
 							error().set("The chararcteristic must be >1",extra);
 							return false;
						}					
						if ( !Zp::setmod(mod) ) {
 							error().set("The chararcteristic must be prime",extra);
 							return false;
						}
						m_results["domain"]=Coeff(Zp(1));
					}
					else {
						error().set(domain.data()+" is not a valid domain",domain);
						return false;
					}
					if (have("epsilon")) {
						cerr<<"the new domain definition abolished epsilon="<<m_results["epsilon"]<<'\n';
						m_results.erase("epsilon");
					}
					if (have("zero")) {
						cerr<<"the new domain definition abolished zero test="<<m_results["zero"]<<'\n';
						m_results.erase("zero");
					}
				}
				else {
					error().set(string("Domain identifier expected but got '")+domain.data()+"'",domain);
					return false;
				}
			}
			else
			if ( commData=="zero" ) {
				symb("=");
				SlocString zeroScheme=name();
				if (zeroScheme) {
					string scheme=lowercase(zeroScheme.data());
					if (scheme=="exact") m_results["zero"]=AlmostEqual::Exact;
					else
					if (scheme=="abs") m_results["zero"]=AlmostEqual::Abs;
					else
					if (scheme=="rel") m_results["zero"]=AlmostEqual::Rel;
					else {
						error().set(zeroScheme.data()+" is not a zero test scheme",zeroScheme);
						return false;
					}			
				}
				else {
					error().set(string("Zero test scheme expected but got '")+zeroScheme.data()+"'",zeroScheme);
					return false;
				}
			}
			else
			if ( commData=="epsilon" ) {
				if (have("domain")) {
					symb("=");
					Coeff domain=m_results["domain"];
					IntExprParser<Coeff> coeffParser( scanner(),error(),domain.exact(), domain.zero(),domain.unit());
					Coeff eps;
					if (!coeffParser.expr(eps)) return false;
					if (eps<domain.zero()) {
						error().set("epsilon must be positive",command);
						return false;
					}
					m_results["epsilon"]=eps;
        }
				else {
					error().set("domain not specified",command);
					return false;
				}
			}
			else
			if ( commData=="ebound" ) {
 				symb("=");
 				IntExprParser<int> intParser( scanner(),error(),true, 0,1);
 				int eb;
 				if (!intParser.expr(eb)) return false;
 				if (eb<0) {
 					error().set("E-bound must be positive",command);
 					return false;
 				}
 				m_results["ebound"]=eb;
			}
			else
			if ( commData=="mm" || commData=="maple" ) {
				symb("=");
				SlocString yn=name();
				if (yn) {
					string answer=lowercase(yn.data());
					if (answer=="yes" || answer=="no") {
						m_results[commData]=answer;
					}
					else {
						error().set("Just say 'yes' or 'no'",yn);
						return false;
					}
				}
				else {
					error().set("Computation option (yes/no) expected",yn);
					return false;
				}
      }
			else
			if ( commData=="report" ) {
				symb("=");
				string fname;
        SlocString pname;
        for(;;) {
          pname=name();
          if (!pname) pname=number();
          if (!pname) pname=symb(".\\/:");
          if (pname) fname+=pname.data();
          else break;
        }
        if (!fname.empty()) {
					SlocString comma=symb(",");
					int printLevel=0;
					if (comma) {
			 			IntExprParser<int> pp( scanner(),error(),true, 0,1);							
 						if (!pp.expr(printLevel)) return false;
					}
					else {
						printLevel=1;
						cerr<<"No report printLevel specified, using "<<printLevel<<" by default.\n";
					}
					comma=symb(",");
					int matrWidth=0;
					if (comma) {
			 			IntExprParser<int> mw( scanner(),error(),true, 0,1);							
 						if (!mw.expr(matrWidth)) return false;
					}
					else {
						matrWidth=180;
						cerr<<"No report matrix output width specified, using "<<matrWidth<<" by default.\n";
					}
          Report r=Report(fname.c_str(),printLevel,matrWidth);
          if (!r.rs()) {
            error().set("the report file is not writable",pname);
            return false;
          }
          m_results["report"]=r;
				}
				else {
					error().set("no report file name specified",pname);
					return false;
				}
			}
			else
			if ( commData=="ideal" ) {
				if ( !have("alphabet") ) {
					error().set("Alphabet not specified",command);
					return false;
				}
				if ( !have("domain") ) {
					error().set("Domain not specified",command);
					return false;
				}
				if ( !have("termorder") ) {
					error().set("TermOrder not specified",command);
					return false;
				}
				if ( !have("etermorder") ) {
				  m_results["etermorder"]=EtermOrder::te();
					cerr<<"Term order extension not specified, using '"<<m_results["etermorder"]<<"' by default\n";
				}
				if ( !have("zero") ) {
					m_results["zero"]= AlmostEqual::guessScheme(m_results["domain"]);
					cerr<<"Zero test scheme not specified, using '"<<AlmostEqual::schemeName(m_results["zero"])<<"' by default\n";
				}
				if ( !have("zero") ) {
					m_results["zero"]= AlmostEqual::guessScheme(m_results["domain"]);
					cerr<<"Zero test scheme not specified, using '"<<AlmostEqual::schemeName(m_results["zero"])<<"' by default\n";
				}
				if ( !have("ebound") ) {
					m_results["ebound"]= 3;
					cerr<<"No E-bound specified, using '"<<m_results["ebound"]<<"' by default\n";
				}
				if ( !have("epsilon") ) {
					Coeff domain=m_results["domain"];
					bool exact = (AlmostEqual::Scheme)m_results["zero"]==AlmostEqual::Exact;
          if (exact) {
            m_results["epsilon"]= domain.unit();
          }
          else {
            int mlen=12;
            if (domain.isType<float>()) mlen=23;
            else
            if (domain.isType<double>()) mlen=52;
            else
            if (domain.isType<Fpoint>()) mlen=Fpoint::precision();
						int hm = mlen/int(m_results["ebound"]);
            m_results["epsilon"]= ( domain.unit()+domain.unit() ).pow(-hm);
            cerr<<"No epsilon specified, using '"<<m_results["epsilon"]<<"(=2^-"<<hm<< ")' by default\n";
          }
				}
				
				if ( !have("report") ) {
					m_results["report"]= Report("/dev/tty",1,180);
					cerr<<"No report file specified, using'"<<m_results["report"]<<"' by default\n";
				}
				if (have("digits")) {
					Report r=m_results["report"];
					r.rs()<<std::setprecision(int(m_results["digits"]));
				}
				Coeff eps=m_results["epsilon"];
				AlmostEqual::Scheme scm=m_results["zero"];
				m_results["context"] = PolyContext(	m_results["alphabet"],
 																						m_results["termorder"],
 																						m_results["etermorder"],
 																						m_results["domain"],
 																						AlmostEqual(eps,scm),
 																						m_results["ebound"]
 																					);
				symb("=");
        SlocString oBracket=symb("{");
				Polins polins;
				for (;;) {
					Coeff domain=m_results["domain"];
					IntPolinParser polinParser( scanner(),error(),m_results["context"]);
					Polin pol;
					if (!polinParser.expr(pol)) return false;
  				if (pol.isZero()) {
  					error().set("zero polynomial not allowed",oBracket);
  					return false;
  				}
					polins.push_back(pol);
					if ( !symb(",") ) break;
				}
				if ( oBracket ) {
					SlocString cBracket=symb("}");
					if (!cBracket) error().set("no }",cBracket);
					return false;
				}
				m_results["ideal"]=polins;				
			}
			else {
				error().set(string("unknown command ")+commData,command );
				return false;
			}
  	}
		else {
			error().set("command expected",command);
			return false;
		}
		SlocString semi=symb(";");
		if (!semi) {
			error().set("; delimiter expected",semi);
			return false;
		}
		if ( end() ) {
			if ( !have("mm") ) {
				m_results["mm"]=string("yes");
			}
			if ( !have("maple") ) {
				m_results["maple"]=string("yes");
			}
			if ( !have("ideal") ) {
				error().set("The ideal generators not specified",command);
				return false;
			}
			//cout<<"Read from the input:\n"<<m_results["context"]<<m_results["ideal"];
			return true;
		}
	}
}
