#include <iostream>
#include "integer.h"

using namespace std;

const char * const Integer_error::prefix="MP integer error: ";

void Integer::write(ostream& os, const mpz_t src) {
int base=10;
 	const char *base_prefix="";
 	switch (os.flags() & ios::basefield) {
 		case ios::hex:
    	base=16;
    	base_prefix="0x";
    	break;
 		case ios::oct:
    	base=8;
    	base_prefix="0";
    	break;
    default:
      break;
 	}

 	const unsigned size=mpz_sizeinbase(src, base)+2;
 	char text[size], *t=text;
 	mpz_get_str(text, base, src);

 	if (os.flags() & ios::showbase) {
    if (*t != '0') {
      if (*t == '-') os.put(*t++);
      os << base_prefix;
    }
 	}
 	os << t;
}

ostream &Integer::print(ostream &os) const {
  write(os,rep);
  return os;
}

void Integer::read(istream& is, mpz_t dst, char delim, bool allow_sign)  noexcept(false) {
  string text;
  char c;
  if (allow_sign) {
		is >> c;
		switch (c) {
      case '-':
				text += c;  // NOBREAK
      case '+':
				is >> c;  // there may be some spaces after the sign, too
		}
  } 
  else {
		is.get(c);
  }
  if (is.eof()) {
		throw Integer_error("unexpexted end of read stream");
  }

  int base=0;    // per default mpz_set_str must detect the numeric base automatically
  switch (is.flags() & ios::basefield) {
   	case ios::hex:
      base=16; break;
   	case ios::oct:
      base=8;  break;
   	case ios::dec:
      base=10; break;
    default:
      break;
  }

  // gather all characters up to the first whitespace, delimiter, or end of file
  int cnt=0;
  while (!(is.eof() || isspace(c) || c==delim)) {
		text += c;
    is.get(c);
    ++cnt;
  }
  if (!is.eof()) is.unget();

  if (!cnt || mpz_set_str(dst, text.c_str(), base)) throw Integer_error("bad read stream format");
}

istream &Integer::scan(istream &is) noexcept(false) {
  try {
		read(is,rep);
  }
  catch(Integer_error &) {
    is.setstate(ios::failbit);
    throw;
  }
  return is;
}

struct gmp_alloc_init {
  gmp_alloc_init();
};

gmp_alloc_init::gmp_alloc_init() {
  //mp_set_memory_functions (std::__default_alloc_template<true,0>::allocate, std::__default_alloc_template<true,0>::reallocate, std::__default_alloc_template<true,0>::deallocate);
}

static gmp_alloc_init do_init;
