#if !defined(__MONOM_H__)
#define __MONOM_H__

#include <iostream>
#include "checks.h"
#include "shared_object.h"
#include "eterm.h"
#include "coeff.h"

class Polin;
class Erow;

class Monom {

	friend class Polin;
	friend class Erow;

	shared_object<Coeff> m_coeff;
	shared_object<Eterm> m_eterm;

	Coeff &coeff() {return *m_coeff;}
	Eterm &eterm() {return *m_eterm;}

	Monom(const Coeff &c, const shared_object<Eterm> &et):m_coeff(c),m_eterm(et) {}
	Monom(const shared_object<Coeff> &c, const Eterm &et):m_coeff(c),m_eterm(et) {}

public:

	Monom():m_coeff(PolyContext::defaultContext.coeff().unit()),m_eterm(Eterm(PolyContext::defaultContext)) {}
	Monom(const PolyContext &ctx):m_coeff(ctx.coeff().unit()),m_eterm(Eterm(ctx)) {}
	Monom(const Coeff &c, const Eterm &et):m_coeff(c),m_eterm(et) {}
	Monom(const Coeff &c, const Term &t, int ed=0):m_coeff(c),m_eterm(Eterm(t,ed)) {}
	Monom(const PolyContext &ctx, const Coeff &c):m_coeff(c),m_eterm(Eterm(ctx)) {}
	Monom(const Eterm &et):m_coeff(et.context().coeff().unit()),m_eterm(et) {}
	Monom(const Term &t):m_coeff(t.context().coeff().unit()),m_eterm(t) {}

	Monom(const Monom &m):m_coeff(m.m_coeff),m_eterm(m.m_eterm) {}

	template<class Iter>
	Monom(const PolyContext &ctx, const Coeff &c, int ed, Iter iter):m_coeff(c),m_eterm(ctx,ed,iter) {}

  const Coeff &coeff() const {return *m_coeff;}
  const Eterm &eterm() const {return *m_eterm;}

	const PolyContext &context() const {return eterm().context();}
	
	Monom &operator=(const Monom &m) {
		m_coeff = m.m_coeff;
		m_eterm = m.m_eterm;
		return *this;
	}
	
	bool operator==(const Monom &m) const {
		return eterm()==m.eterm() && coeff()==m.coeff();
	}

	bool operator!=(const Monom &m) const {
		return eterm()!=m.eterm() || coeff()!=m.coeff();
	}

	bool operator<(const Monom &m) const {
		return eterm()<m.eterm() || (eterm()==m.eterm() && coeff()<m.coeff());
	}
	bool isUnit() const {return coeff().isUnit() && eterm().isUnit();}
	bool isZero() const {return eterm().isZero() || coeff().isZero();}

	Monom unit() const {return Monom(context());}

	int degree() const {return eterm().degree();}
	int tdegree() const {return eterm().tdegree();}
	int edegree() const {return eterm().edegree();}

	void setEdegree(int ed) {
		eterm().setEdegree(ed);
	}
	void resetEdegree() {
		eterm().resetEdegree();
	}

	Monom operator*(const Monom &m) const;
	Monom &operator*=(const Monom &m);

	Monom operator*(const Coeff &c) const;
	Monom &operator*=(const Coeff &c);

	Monom operator*(const Eterm &et) const;
	Monom &operator*=(const Eterm &et);

	Monom operator/(const Coeff &c) const;
	Monom &operator/=(const Coeff &c);

	Monom operator+(const Monom &m) const;
	Monom &operator+=(const Monom &m);

	Monom operator-(const Monom &m) const;
	Monom &operator-=(const Monom &m);

	Monom operator-() const {
		return Monom(-coeff(),m_eterm);
	}

	void print(std::ostream &os) const;
};

inline std::ostream &operator<<(std::ostream &os,const Monom &m) {
	m.print(os);
	return os;
}


#endif //__MONOM_H__
