#include <algorithm>
#include <iterator>
#include <functional>
#include <map>
#include <iomanip>
#include "string_cast.h"
#include "multmatr.h"

using namespace std;

MultMatr::MultMatr(const SetOfTerms &ns):Inherited(ns.size()),m_ns(ns),m_var("...") {
	for_each(begin(),end(),bind2nd(mem_fun_ref(&Row::resize),size()));
	clear();
}


void MultMatr::clear() {
	if (m_ns.empty()) return;
	Coeff zero=m_ns.begin()->context().coeff().zero();
	for(iterator row=begin();row!=end();++row) {
		fill(row->begin(),row->end(),zero);
	}
}

void MultMatr::operator()(const BorderBasis &bb,int v) {
	if (m_ns.empty()) return;
	const PolyContext &ctx=m_ns.begin()->context();
	Coeff zero=ctx.coeff().zero();
	Coeff unit=ctx.coeff().unit();
	m_var=ctx.abc()[v];
	iterator row=begin();
	for(SetOfTerms::const_iterator t=m_ns.begin();t!=m_ns.end();++t,++row) {
		Term s=*t*t->var(v);
		SetOfTerms::const_iterator w=m_ns.find(s);
		if (w!=m_ns.end()) {
			Row::iterator col=transform(m_ns.begin(),w,row->begin(),__gnu_cxx::constant_unary_fun<Coeff,Term>(zero));
			*col++=unit;
			col=transform(++w,m_ns.end(),col,__gnu_cxx::constant_unary_fun<Coeff,Term>(zero));
			PRECONDITION(col==row->end());
		}
		else {
			BorderBasis::const_iterator b=bb.find(s);
			PRECONDITION(b!=bb.end());
			Row::iterator col=row->begin();
			for(SetOfTerms::const_iterator j=m_ns.begin();j!=m_ns.end();++j,++col) {
				Polin::const_iterator m=b->second.monom(*j);
				*col=(m==b->second.end() ? zero:-m->coeff());
			}
		}
	}
}
void MultMatr::print(std::ostream &os) const {
	if (empty()) {
		os<<"<empty>\n";
		return;
	}
	vector<size_t> tabs(m_ns.size());
	transform(m_ns.begin(),m_ns.end(),tabs.begin(),compose1(mem_fun_ref<string::size_type,string>(&string::size),string_caster<Term>()));
	size_t tlen=*max_element(tabs.begin(),tabs.end())+m_var.size()+3;
	for(const_iterator row=begin();row!=end();++row) {
		vector<size_t>::iterator j=tabs.begin();
		for(Row::const_iterator col=row->begin();col!=row->end();++col,++j) {
			*j=max(*j,string_cast(*col,os.precision()).size());
		}
	}
  size_t total=accumulate(tabs.begin(),tabs.end(),tabs.size());
  fill_n(ostream_iterator<char>(os),tlen,' ');
	os<<'|';
	vector<size_t>::iterator j=tabs.begin();
	for (SetOfTerms::const_iterator t=m_ns.begin();t!=m_ns.end();++t,++j) {
		os<<setw(*j)<<string_cast(*t)<<' ';
	}
	os<<'\n';
  fill_n(ostream_iterator<char>(os),tlen,'-');
	os<<'+';
	fill_n(ostream_iterator<char>(os),total,'-');
	os<<'\n';
  SetOfTerms::const_iterator t=m_ns.begin();
	for(const_iterator row=begin();row!=end();++row,++t) {
		string rowName=string_cast(*t)+" * "+m_var;
		os<<setw(tlen)<<rowName<<'|';
		j=tabs.begin();
		for(Row::const_iterator col=row->begin();col!=row->end();++col,++j) {
			os<<setw(*j)<<string_cast(*col,os.precision())<<' ';
		}
		os<<'\n';
	}
}

MultMatr &MultMatr::comb(const MultMatr &mm, const Coeff &c) {
	PRECONDITION(m_ns==mm.m_ns);
	iterator row1=begin();
	const_iterator row2=mm.begin();
	for(;row1!=end();++row1,++row2) {
		Row::iterator col1=row1->begin();
		Row::const_iterator col2=row2->begin();
		for(;col1!=row1->end();++col1,++col2) {
			*col1+=*col2*c;
		}
	}
	return *this;
}


void MultMatr::printMaple(std::ostream &os) const {
	os<<'[';
	const_iterator first=begin();
	const_iterator last=end();
	if (first!=last) {
		--last;
		for(;first!=last;++first) {
			printMapleRow(os,first);
			os<<",\n";
		}
		printMapleRow(os,first);
	}
	os<<"]\n";
}

void MultMatr::printMapleRow(std::ostream &os, const_iterator row) {
	os<<'[';
	Row::const_iterator first=row->begin();
	Row::const_iterator last=row->end();
	PRECONDITION(first!=last);
	copy(first,--last,ostream_iterator<Coeff>(os,","));
	os<<*last;
	os<<']';
}
