#include "polinparser.h"


bool IntPolinParser::atom(Polin &val) {
	SlocString n=number(exact());
	if (n) {
		Coeff c=context().coeff().zero();
		stringScan(n.data(),c);
		val=Polin(Monom(c,Term(context()).unit()));
	}
	else {
		SlocString var=name();
		if (var) {
			int varInd=context().abc().at(var.data());
			if (varInd<0) {
				error().set(string("No variable ")+var.data()+" in the alphabet",var);
				return false;
			}
			val=Polin(Monom(Eterm(Term(context()).var(varInd))));
			//cerr<<"var = "<<val<<'\n';
		}
		else {
  		SlocString obrace=symb("(");
  		if (!obrace) {
				error().set("No (",obrace);
				return false;
			}
  		bool se=sum(val);
  		if (!se) return false;
  		SlocString cbrace=symb(")");
  		if (!cbrace) {
				error().set("No )",cbrace);
				return false;
			}
		}
	}
	return true;
}

bool IntPolinParser::exp(Polin &val) {
	bool a=atom(val);
	if (!a) return false;
	SlocString sign=symb("^");
	if (sign) {
		int p=0;
		IntExprParser<int> powParser( scanner(),error(),true );
		a=powParser.atom(p);
		if (!a) return false;
		if (p<0) {
			if (val.size()==1 && !val.lc().isZero() && val.lt().isUnit()) {
				val=Polin(Monom( val.lc().pow(p), val.lt() ));
			}
			else {
				error().set("negative power of a polynomial not defined ",sign);
				return false;
			}
		}
		else
		if (p>0)
			val=val.power(p);
		else
			val=unit();
		//cerr<<val<<'\n';
	}
	return true;
}

bool IntPolinParser::mul(Polin &val) {
	bool first=exp(val);
	if (!first) {
		return false;
	}
	for(;;) {
		SlocString sign=symb("*");
		if ( sign ) {
			Polin val1=zero();
			bool next=exp(val1);
			if (!next) return false;
			//cerr<<val<<" * "<<val1<<'=';
			val=val*val1;
			//cerr<<val<<'\n';
		}
		else
 		if ( (sign=symb("/")) ) {
 			Coeff c=context().coeff().zero();
 			IntExprParser<Coeff> coeffParser( scanner(),error(),exact(), context().coeff().zero(),context().coeff().unit());
 			bool dividend=coeffParser.atom(c);
 			if (!dividend) return false;
 			if (c.isZero()) {
 				error().set("can not divide by zero",sign);
 				return false;
 			}
 			val=val/c;
 		}
 		else break;
	}
	return true;
}

bool IntPolinParser::sum(Polin &val) {
	SlocString sign=symb("+-");
	bool first=mul(val);
	if (!first) return false;
	if ( selSymb(sign)=='-' ) val=-val;
	for(;;) {
		sign=symb("+-");
		if (!sign) break;
		Polin val1=zero();
		bool next=mul(val1);
		if (!next) return false;
		if ( selSymb(sign)=='-' ) val=val-val1;else val=val+val1;
	}
	return true;
}
