#include <sstream>
#include <iomanip>
#include "alt_binary_compose.h"
#include "printematrix.h"

using namespace std;

Ptable::Ptable(const SetOfEterms &st) {
	for (SetOfEterms::const_iterator et=st.begin();et!=st.end();++et) {
		m_tab.insert(Tab::value_type(*et,0));
	}
	index_type count=0;
	for (Tab::iterator i=m_tab.begin();i!=m_tab.end();++i) i->second=count++;
};

const Eterm &Ptable::maximal() const {
	Tab::const_iterator result=max_element(m_tab.begin(),m_tab.end(),alt_compose2(less<Tab::data_type>(),__gnu_cxx::select2nd<Tab::value_type>(),__gnu_cxx::select2nd<Tab::value_type>()));
	PRECONDITION(result!=m_tab.end());
	return result->first;
}

bool ColLess::operator()(const Eterm &et1,const Eterm &et2) const {
	bool red1=m_tab->contains(et1);
	bool red2=m_tab->contains(et2);
	if (red1) {
		if (red2) return m_tab->isBefore(et1,et2);
		return true;
	}
	if (red2) return false;
	return et2<et1;
}

ostream &operator<<(ostream &os,const Row &row) {
	for(Row::const_iterator i=row.begin();i!=row.end();++i) {
		os<<' '<<i->first<<" -> "<<i->second;
	}
	return os;
}

bool RowLess::operator()(const Row &r1,const Row &r2) const {
	Row::const_iterator first1=r1.begin();
	Row::const_iterator  last1=r1.end();
	Row::const_iterator first2=r2.begin();
	Row::const_iterator  last2=r2.end();
	for (;first1!=last1 && first2!=last2;++first1,++first2) {
		if ( m_colComp(first1->first,first2->first) ) return true;
		if ( m_colComp(first2->first,first1->first) ) return false;
		if ( first1->second<first2->second ) return true;
		if ( first2->second<first1->second ) return false;
	}
	return first1==last1 && first2!=last2;
}

void Row::tuneTabs(int &tagLen,Tabs &tabs) const {
	for(const_iterator i=begin();i!=end();++i) {
		const Eterm &et=i->first;
		tabs[et]=max(tabs[et],int(i->second.length()));
	}
	tagLen=max(tagLen,int(tag().length()));
}

Row::Row(const Tpolin &bp, const ColLess &cl):Inherited(cl) {
	Polin::const_iterator first=bp.begin();
	Polin::const_iterator  last=bp.end();
	for (;first!=last;++first) {
		ostringstream os;
		os<<first->coeff();
		(*this)[first->eterm()]=os.str();			
	}
	ostringstream os;
	os<<bp.tag();
	m_tag=os.str();
}

Row::Row(Tabs &tabs):Inherited(tabs.key_comp()) {
	for (Tabs::const_iterator i=tabs.begin();i!=tabs.end();++i) {
		ostringstream os;
		os<<i->first;
		insert(value_type(i->first,os.str()));
	}
	int tagLen=0;
	tuneTabs(tagLen,tabs);
}

Tabs::Tabs(const Rows &rows):Inherited(rows.key_comp().colLess()) {}

Tabs::const_iterator Tabs::firstIrred() const {
	const_iterator result=end();
	const Ptable &ptable=key_comp().ptable();
	if (!ptable.empty()) {
		const Eterm &et=ptable.maximal();
		result=upper_bound(et);
	}
	return result;
}

pair<Tabs::const_iterator,int> PrintEmatrix::printRow(ostream &os, Tabs::const_iterator watershed, const Row &row, Tabs::const_iterator first,Tabs::const_iterator last,int pageWidth) {
	int printed=0;
	Row::const_iterator cell=row.lower_bound(first->first);
	if (first!=last) {
  	
  	while(first!=last)	{
  		if (first==watershed) {
  			os<<" |";
  			printed+=2;
  			pageWidth-=2;
  		}
  		
  		int colWidth=first->second+1;
			
  		if (cell!=row.end() && cell->first==first->first) {
				os<<setw(colWidth)<<cell->second.c_str();
				++cell;
			}
			else {
  			fill_n(ostream_iterator<char>(os),colWidth-1,' ');
  			os<<'.';
			}
  		printed+=colWidth;
  		pageWidth-=colWidth;
  		++first;
  		if (pageWidth<=0) break;
  	}
	}
	return pair<Tabs::const_iterator,int>(first,printed);
}

ostream &PrintEmatrix::operator()(ostream &os, const Elimination::Ematrix &em,int pageWidth) {
	m_rows.clear();
	m_tabs.clear();
	if (em.empty()) {
		os<<"empty>\n";return os;
	}
	int tagLen=0;
	for(Elimination::ematrix_iterator bpi=em.begin();bpi!=em.end();++bpi) {
		Rows::const_iterator row=m_rows.insert(Row(*bpi,m_rows.key_comp().colLess()));
		row->tuneTabs(tagLen,m_tabs);
		//cout<<"Inserted: "<<*row<<'\n';
	}
	Row colNames(m_tabs);
	//cout<<"Inserted names: "<<colNames<<'\n';
	Tabs::const_iterator firstTab=m_tabs.begin();
	Tabs::const_iterator  lastTab=m_tabs.end();
	int width=pageWidth-tagLen-2;
	Tabs::const_iterator watershed=m_tabs.firstIrred();
	while (firstTab!=lastTab) {
		fill_n(ostream_iterator<char>(os),tagLen+1,' ');
		os<<'|';
		pair<Tabs::const_iterator,int> advance=printRow(os,watershed,colNames,firstTab,lastTab,width);
		Tabs::const_iterator nextTab=advance.first;
		int tableSize=advance.second;
		os<<'\n';
		fill_n(ostream_iterator<char>(os),tagLen+1,'-');
		os<<'+';
		fill_n(ostream_iterator<char>(os),tableSize,'-');
		os<<'\n';
		for(Rows::const_iterator i=m_rows.begin();i!=m_rows.end();++i)	{
	  	os<<setw(tagLen+1)<<i->tag().c_str();
			os<<'|';
			printRow(os,watershed,*i,firstTab,lastTab,width);
			os<<'\n';
		}
		firstTab=nextTab;
	}
	return os;
}
