#include <cmath> 
#include <cstring>
#include <sstream>
#include "rational.h"

using namespace std;

const char * const Rational_error::prefix="MP rational error: ";

Rational Rational::pow(long k) const {
 	Integer my_num=num();
 	Integer my_den=denom();
 	if (k>0) {
 		return Rational(my_num.pow(k),my_den.pow(k));
 	}
 	else {
 		PRECONDITION(!my_num.isZero());k=-k;
 		return Rational(my_den.pow(k),my_num.pow(k));
 	}
}

Rational& Rational::set(const char* s) noexcept(false) {
  const char* digit(s);
  while (*digit && *digit!='/') ++digit;
  int numerator_digits(digit-s);
  if (!numerator_digits)
      throw Rational_error("syntax error in string");

  char numerator[numerator_digits+1];
  strncpy(numerator,s,numerator_digits);
  numerator[numerator_digits]=0;
  if (mpz_set_str(mpq_numref(rep), numerator, 0) < 0)
	throw Rational_error("syntax error in numerator");

  if (*digit) {
    ++digit;
    if (*digit<'0' || *digit>'9' || mpz_set_str(mpq_denref(rep), digit, 0) < 0)
	  throw Rational_error("syntax error in denominator");
		mpq_canonicalize(rep);
  } 
  else {
		mpz_set_ui(mpq_denref(rep),1);
  }
  return *this;
}

Rational& Rational::operator= (double d) {
  double x=d;
  MP_INT *num=mpq_numref(rep),
         *den=mpq_denref(rep);
  mpz_set_ui(num, 0);
  mpz_set_ui(den, 1);
  if (x != 0.0) {
 		int neg = x < 0;
    if (neg) x = -x;
      
  	const long shift = 15;         // a safe shift per step
  	const double width = 32768.0;  // = 2^shift
  	const int maxiter = 20;        // ought not be necessary, but just in case,
                                     // max 300 bits of precision
  	int expt;
  	double mantissa = frexp(x, &expt);
  	long exponent = expt;
  	double intpart;
  	int k = 0;
  	while (mantissa != 0.0 && k++ < maxiter) {
    	mantissa *= width;
    	mantissa = modf(mantissa, &intpart);
    	mpz_mul_2exp(num, num, shift);         // num <<= shift;
    	mpz_add_ui(num, num, static_cast<long>(intpart));   // num += (long)intpart;
    	exponent -= shift;
  	}
  	if (exponent > 0)  mpz_mul_2exp(num, num, exponent); //num <<= exponent;
  	else
  	if (exponent < 0)  mpz_mul_2exp(den, den, -exponent); //den <<= -exponent;
  	if (neg) mpz_neg(num, num);
  }
  mpq_canonicalize(rep);
  return *this;
}


ostream &Rational::print(ostream &os) const {
  Integer::write(os,mpq_numref(rep));
  if (mpz_cmp_ui(mpq_denref(rep),1)) { os.put('/'); Integer::write(os,mpq_denref(rep)); }
  return os;
}

istream &Rational::scan(istream &is) {
  string text;
  char c;
  is>>c;
  bool fp=false;
  while (!(is.eof() || isspace(c) || c=='\0')) {
    if (strchr("eE.",c)) fp=true;
		text += c;
    is.get(c);
  }
  if (!is.eof()) is.unget();
  std::istringstream ss(text);
  if (fp) {
		double d;
		ss>>d;
		*this=d;
  }
  else {
		Integer::read(ss,mpq_numref(rep),'/');
		if (ss.peek() == '/') {
	 		ss.ignore();
	 		Integer::read(ss,mpq_denref(rep),0,false);
	 		mpq_canonicalize(rep);
		}
		else {
	  	mpz_set_ui(mpq_denref(rep), 1);
		}
  }
  return is;
}
