#include "reduction.h"

void Reduction::mutual(Basis &b) {
	REPORT(1,"Entering auto-reduction of the basis.\n");
	for (;;) {
		m_symbred.mutual(b);
		if ( m_symbred.red().empty() && m_symbred.keep().empty() ) {
			REPORT(1,"No reducible terms found - exit reduction.\n");
			REPORT(3,"The auto-reduced set of polynomials is\n"<<b<<'\n');
			break;
		}
		m_elim(b,m_symbred.red(),m_symbred.keep(),false);
	}
}

void Reduction::mutualG(Basis &b) {
	REPORT(1,"Entering auto-reduction of a Groebner basis.\n");
	REPORT(1,"Constructing a minimal Groebner basis.\n");
	for (Basis::const_iterator i=b.begin();i!=b.end();++i) {
		for (Basis::const_iterator j=b.begin();j!=b.end();) {
			if (i==j) {
				++j;
				continue;
			}
			if ( i->lt().divides(j->lt()) ) {
				REPORT(3,"Removing redundant "<<j->tag()<<'\n');
				REPORT(4,"which is "<<*j<<'\n');
				b.erase(j++);
			}
			else
				++j;
		}
	}
  REPORT(1,"The minimal Groebner basis with "<<b.size()<<" polynomials is\n"<<b<<'\n');

  for (Basis::const_iterator p=b.begin();p!=b.end();)  {
    Polin f=*p;
    for(size_t i=1;i<f.size();) {
      const Eterm s=f.monom(i).eterm();
      Basis::div_iterator first=b.lower_divisor(s);
      Basis::div_iterator last=b.upper_divisor(s);
      if (first!=last) {
        Basis::const_iterator q=first;
        while (++first!=last) {
          if (q->lc().abs()<first->lc().abs()) q=first;
        }
        const Coeff c=-f.monom(i).coeff()/q->lc();
        f=f.eliminate(*q*(s.divide(q->lt())),c,s);
      }
      else {
        ++i;
      }
    }
    b.insert(p,Tpolin(f,p->tag()));
    b.erase(p++);
  }
  
#if 0  
  m_symbred.mutual(b);
  if ( !m_symbred.red().empty()  || !m_symbred.keep().empty() )  {
      m_elim(b,m_symbred.red(),m_symbred.keep(),false);
  }
  else {
    REPORT(1,"No reducible terms found - the minimal Groebner basis appears to be reduced.\n");
  }
  for (;;) {
    m_symbred.mutual(b);
    if ( m_symbred.red().empty()  && m_symbred.keep().empty() )  break;
    m_elim(b,m_symbred.red(),m_symbred.keep(),false);
  }
#endif
}

void Reduction::operator()(Basis &b,const Basis &gb) {
	REPORT(1,"Entering reduction of the set of polynomials.\n");
	m_simp.clear();
	for (;;) {
		m_symbred(b,gb,m_simp);
		if ( m_symbred.red().empty() && m_symbred.keep().empty() ) {
			REPORT(1,"No reducible terms found - exit reduction.\n");
			REPORT(3,"The reduced set of polynomials is\n"<<b<<'\n');
			break;
		}
		m_elim(b,m_symbred.red(),m_symbred.keep(),true);
		m_simp.insert(m_elim.simp().begin(),m_elim.simp().end());
	}
}

void Reduction::spairs(Basis &b,const Basis &gb) {
	REPORT(1,"Entering reduction of the set of s-pairs.\n");
	m_simp.clear();
	m_symbred.spairs(b,gb,m_simp);
	for (;;) {
		m_elim(b,m_symbred.red(),m_symbred.keep(),true);
		m_simp.insert(m_elim.simp().begin(),m_elim.simp().end());
		m_symbred(b,gb,m_simp);
		if ( m_symbred.red().empty() && m_symbred.keep().empty() ) {
			REPORT(1,"No reducible terms found - exit reduction.\n");
			REPORT(3,"The reduced set of polynomials is\n"<<b<<'\n');
			break;
		}
	}
}

void Reduction::clear() {
	m_symbred.clear();
	m_elim.clear();
	m_simp.clear();
}
