#include <cctype>
#include <cstring>
#include "scanner.h"

using namespace std;

ostream &LocString::print(ostream &os) const {
	return os<<m_data<<", L:"<<m_line<<", C:"<<m_col;
}

const char * Token::typeName(Type t) {
	switch (t) {
		case Symbol:return "symbol";
		case Digits:return "digits";
		case Name  :return "name";
		case End	 :return "end";
		default    :return "ERROR";
	}
}

ostream &Token::print(ostream &os) const {
	return os<<"Token("<<typeName(m_type)<<", "<<locString()<<')';
}

void Scanner::skipWhite() {
	static const char * const white=" \t\n\r";
	bool comment=false;
	while (!m_text.eot()) {
		char c=m_text.getc();
		if (c=='%') comment=true;
		if (c=='\n') comment=false;
		if (comment) continue;
		if ( strchr(white,c) ) continue;
		m_text.ungetc();break;
	}
}

Token Scanner::get() {
	skipWhite();
	int l,c;
	m_text.getpos(l,c);
	if ( m_text.eot() ) return Token("",l,c,Token::End);
 	char symb=m_text.getc();
 	if ( isalpha(symb) ) {
 		string s;
 		for(;;) {
 			s+=symb;
 			if (m_text.eot()) break;
 			symb=m_text.getc();
 			if (isalnum(symb)) continue;
 			m_text.ungetc();break;
 		}
 		return Token(s,l,c,Token::Name);
 	}
 	else
 	if ( isdigit(symb) ) {
 		string s;
 		for(;;) {
 			s+=symb;
 			if (m_text.eot()) break;
 			symb=m_text.getc();
 			if (isdigit(symb)) continue;
 			m_text.ungetc();break;
 		}
 		return Token(s,l,c,Token::Digits);
 	}
	return Token(string(&symb,1),l,c,Token::Symbol);
}
