#if !defined __SCANNER_H__
#define __SCANNER_H__

#include <iostream>
#include <string>
#include "text.h"

class Scanner;

class LocString {
	std::string m_data;
	int m_line,m_col;
public:
	LocString(const std::string &s,int l,int c):m_data(s),m_line(l),m_col(c) {}
	LocString(const LocString &ls):m_data(ls.m_data),m_line(ls.m_line),m_col(ls.m_col) {}
	LocString &operator=(const LocString &ls) {
		if (this!=&ls) {
			m_data=ls.m_data;
			m_line=ls.m_line;
			m_col=ls.m_col;
		}
		return *this;
	}
	const std::string &data() const {return m_data;}
	int line() const {return m_line;}
	int col() const {return m_col;}

	LocString operator+(const LocString &ls) const {
		return LocString(data()+ls.data(),line(),col());
	}
	LocString &operator+=(const LocString &ls) {
		m_data+=ls.data();return *this;
	}
	std::ostream &print(std::ostream &os) const;
};

inline std::ostream &operator<<(std::ostream &os,const LocString &ls) {
	return ls.print(os);
}

struct Token:public LocString {
	friend class Scanner;
	enum Type {Symbol,Digits,Name,End};
	static const char * typeName(Type);
private:
	Type m_type;
	Token(const std::string &s,int l,int c,Type t):LocString(s,l,c),m_type(t) {}
public:
	Type type() const {return m_type;}
	std::ostream &print(std::ostream &os) const;
	const LocString &locString() const {return *this;}
};

inline std::ostream &operator<<(std::ostream &os,const Token &token) {
	return token.print(os);
}

class Scanner {

	Text m_text;
	void skipWhite();
public:
	Scanner() {}
	Scanner(std::istream &is):m_text(is) {}
	void load(std::istream &is) {m_text.load(is);}
	void reset() {m_text.reset();}
	Token get();
	const std::string &inputLine(int k) const {return m_text.line(k);}
};

#endif //__SCANNER_H__

