#if !defined(__TERM_H__)
#define __TERM_H__

#include <iostream>
#include "shared_object.h"
#include "polcontext.h"
#include "checks.h"

class Term: public shared_array<int> {
	
	typedef shared_array<int> Inherited;
	
	PolyContext m_context;

protected:
	void sameContext(const Term &t) const {
		PRECONDITION(context()==t.context());
	}
public:

	typedef Inherited::size_type size_type;
	typedef Inherited::iterator iterator;
	typedef Inherited::const_iterator const_iterator;
	typedef Inherited::reverse_iterator reverse_iterator;
	typedef Inherited::const_reverse_iterator const_reverse_iterator;

	Term():Inherited(PolyContext::defaultContext.abc().size(),0), m_context(PolyContext::defaultContext) {}
	Term(const PolyContext &ctx):Inherited(ctx.abc().size(),0), m_context(ctx) {}
	Term(const Term &t):Inherited(t),m_context(t.context()) {}
	template<class Iter>
	Term(const PolyContext &ctx, Iter iter):Inherited(ctx.abc().size(),iter), m_context(ctx) {}

	Term &operator=(const Term &t) {
		Inherited::operator=(t);
		m_context = t.context();
		return *this;
	}

	int degree() const {
    return std::accumulate(begin(),end(),0);
  }

	const PolyContext &context() const {return m_context;}

/*	int &operator[](int i) {
		return Inherited::operator[](i);
	}

	int operator[](int i) const {
		return Inherited::operator[](i);
	}*/

	bool operator==(const Term &that) const {
		sameContext(that);
		return context().to().equal(*this,that);
	}

	bool operator<(const Term &that) const {
		sameContext(that);
		return context().to().less(*this,that);
	}
	bool operator>(const Term &that) const {
		return that<*this;
	}

	Term gcd(const Term &t) const;
	Term lcm(const Term &t) const;

	bool isCoprime(const Term &t) const;
	bool isPP(int &var) const;

	bool divides(const Term &t) const;
	Term divide(const Term &t) const;
	bool isUnit() const;	
	static bool isZero() {return false;}
	Term unit() const {return Term(context());}
	Term var(int varInd) const;
	Term varPow(int varInd, int pow) const;

	Term operator*(const Term &t) const;
	Term &operator*=(const Term &t);

	void print(std::ostream &os) const;

};

inline std::ostream &operator<<(std::ostream &os, const Term &t) {
	t.print(os);
	return os;
}

#endif // __TERM_H__
