#if !defined __TERMORDER_H__
#define __TERMORDER_H__

#include <iostream>

class Term;

struct VectorOrder {
	virtual bool less(const Term &t1, const Term &t2) const=0;
	bool equal(const Term &t1, const Term &t2) const;
	virtual void print(std::ostream &os) const=0;
};

struct TermOrder {

	TermOrder(const TermOrder &to):m_order(to.m_order) {}
	TermOrder &operator=(const TermOrder &to) {
		m_order=to.m_order;
		return *this;
	}
	
	static TermOrder lex();	
	static TermOrder revlex();	
	static TermOrder tdeg();	
	static TermOrder revtdeg();	

	bool less(const Term &t1, const Term &t2) const {
		return m_order->less(t1,t2);
	}	
	bool equal(const Term &t1, const Term &t2) const {
		return m_order->equal(t1,t2);
	}
	void print(std::ostream &os) const {
		m_order->print(os);
	}

private:
	TermOrder(VectorOrder * const order):m_order(order) {}

	const VectorOrder * m_order;
};

inline std::ostream &operator<<(std::ostream &os, const TermOrder &to) {
	to.print(os);
	return os;
}

#endif //__TERMORDER_H__
