#if !defined(_UNTYPED_H)
#define _UNTYPED_H

#include <typeinfo>
#include <string>
#include <iostream>
#include "classError.h"

using std::type_info;


class Untyped;

struct Mutant  {

  virtual Mutant *clone() const=0;
  virtual void copy(const Mutant &)=0;
  virtual void print(std::ostream &os) const=0;

  virtual const type_info &type() const=0;
  const char * typeName() const {return type().name();}
  virtual ~Mutant() {}
};


template<class T>
class Tmutant: public Mutant {
  T m_data;
  Tmutant(const T &t):m_data(t) {}
public:
  typedef ClassError<Tmutant<T> > TmutantError;

  Tmutant(); // no body provided, should never be used.
	Tmutant(const Tmutant &); // no body provided, should never be used.

  virtual Mutant *clone() const {return new Tmutant(data());}
  static Mutant *create(const T &t) {return new Tmutant(t);}
  virtual void copy(const Mutant &m) {data()=constRaise(m).data();}


  ~Tmutant() {}

  Tmutant &operator=(const T &t) {
    data()=t;
    return *this;
  }

  virtual void print(std::ostream &os) const {
    os<<data();
  }

  virtual const type_info &type() const {
    return my_type();
  }

  T &data() {return m_data;}
  const T &data() const {return m_data;}

  static Tmutant &raise(Mutant &m) noexcept(false) {
    if ( m.type()!=my_type() ) throw TmutantError(std::string("can not cast to ")+my_name()+" from "+m.typeName()+" sorry..." );
    return (dynamic_cast<Tmutant &>(m));
  }

  static const Tmutant &constRaise(const Mutant &data)  {
    return raise(const_cast<Mutant &>(data));
  }

private:

	static const type_info &my_type() {return typeid(T);}
	static const char *my_name() {return typeid(T).name();}
};

class Untyped {
public:
  typedef ClassError<Untyped> UntypedError;
private:
  Mutant *m_data;

  template<class T> static Tmutant<T> &raise(Mutant *data) noexcept(false){
    if (!data) throw UntypedError("can not up-cast the NULL Mutant");
    return Tmutant<T>::raise(*data);
  }

  template<class T> static const Tmutant<T> &constRaise(const Mutant *data) noexcept(false){
    if (!data) throw UntypedError("can not const up-cast the NULL Mutant");
    return Tmutant<T>::constRaise(*data);
  }

  template<class T> Tmutant<T> &raise() {
    return raise<T>(m_data);
  }

  template<class T> const Tmutant<T> &constRaise() const {
    return constRaise<T>(m_data);
  }

  template<class T> static Mutant * create(const T &t) {return Tmutant<T>::create(t);}	
  static Mutant *clone(const Mutant *data) {return data ? data->clone():0;}

public:

  Untyped():m_data(0) {}
  Untyped(const Untyped &ut): m_data( clone(ut.m_data) ) {}
  template<class T> Untyped(const T &t):m_data( create(t) ) {}

	~Untyped() {if (m_data) delete m_data;}

  Untyped &operator=(const Untyped &ut);
  template<class T> Untyped &operator=(const T &t);

//  template<class T> operator T() {return raise<T>().data();}
  template<class T> operator const T() const {return constRaise<T>().data();}

//  template<class T> T &cast() {return raise<T>().data();}
  template<class T> const T &cast() const {return constRaise<T>().data();}


  void print(std::ostream &os) const;

  const type_info &type() const {
    return m_data ? m_data->type():typeid(void);
  }

  const char *typeName() const {
    return m_data ? m_data->typeName():typeid(void).name();
  }

};

template<class T> Untyped &Untyped::operator=(const T &t) {
	if (m_data) {
		if ( m_data->type()==typeid(T) ) raise<T>()=t;
		else {
			delete m_data;
			m_data=create(t);
		}
	}
	else {
		m_data=create(t);
	}	
	return *this;
}

inline std::ostream &operator<<(std::ostream &os, const Untyped &ut) {
  ut.print(os);
  return os;
}

#endif //_UNTYPED_H
