

#include <stdlib.h>
#include <string.h>
#include <sstream>
#include <numeric>
#include <ext/numeric>
#include "checks.h"
#include "zp.h"

using namespace std;

const char * const Zp_error::prefix="Runtime error in Zp ";

int *Zp::inv=0;
int Zp::p=0;

//long ii=0;

bool Zp::setmod(int m) {
  if (m<2) return false;
  int *itab=new int[m];
  itab[0]=0;
  itab[1]=1;
  for (int i=2;i<m;i++) {
    int r1=abs(m),u1=0,r2=abs(i),u2=1;
    while (r2) {
      int q=r1/r2;
      int r=r1-q*r2;
      int u=u1-q*u2;
      r1=r2;u1=u2;
      r2=r;u2=u;
      //++ii;
    }
    if (r1!=1) {
      delete [] itab;
      return false;
    }
    itab[i]=normalize(u1,m);
  }
  p=m;
  delete [] inv;
  inv=itab;
  return true;
}

void Zp::printInv(ostream &os) {
  for (int i=1;i<p;i++) {
    //os<<' '<<inv[i];
    if ((((unsigned)inv[i]*i)%p)!=1) {
      cout<<"\nCatastrophy "<<i<<':'<<inv[i];
      return;
    }
  }
  os<<'\n';
}

Zp &Zp::set(const char *s) noexcept(false) {
  std::istringstream is(s);
  try {
    scan(is);
  }
  catch(runtime_error &e) {
    throw Zp_error("reading from a string");
  }
  return *this;
}

Zp Zp::pow(long e) const {
	if (e>0) {
  	return __gnu_cxx::power(*this,e);
	}
	else {
		PRECONDITION(!isZero());
		return __gnu_cxx::power(Zp(inv[k],internal),-e);
	}
}

ostream &Zp::print(ostream &os) const {
  os<<toInt();
  return os;
}

istream &Zp::scan(std::istream &is) noexcept(false) {
  int tmp;
  is>>tmp;
  if (!is) throw Zp_error("reading from a stream");
  *this=tmp;
  return is;
}
